import copy
import json

import orjson
import math
import pickle
import time
import tracemalloc
import pypinch
from pypinch._pypinch import dump_bytes as pinch_rust_dump
from pypinch.serialize.serialize import dump_bytes as pinch_py_dump
from pypinch._pypinch import load_bytes as pinch_rust_loads
from pypinch.deserialize.deserialize import load_bytes as pinch_py_loads

# os.environ["MSGPACK_PUREPYTHON"] = "True"
import msgpack

from benchmark_displayer import display_benchmark

# a = 1
# print(load_bytes(pypinch.dump_bytes(a)))
# print(load_bytes(pypinch.dump_bytes(a)) == a)
# print(pypinch.dump_bytes(a))
# print(bytes(dump_bytes(a)))
# exit()
objects = [
    {"a": "sdgaeiogn", "waegw": 123, "sdagweg": list(range(10)), "aegsag": {"asdg": 235, "Asg": b"asg"}},
    {1: "afdbda", "ar": "23wesd", False: 23453, 1234: 12324356, "": {"sgdfn32rwefsdvre": 34}},

    {"a": "sdgaeiogn", "content": b"1243567" * 1024 * 1024 * 50, "sdagweg": list(range(10)), "aegsag": {"asdg": 235, "Asg": b"asg"}},
    [True, False, False] * 1000,
]
objects.clear()

# recursive_obj = {"a": "sdgaeiogn", "waegw": 123, "sdagweg": list(range(10)), "aegsag": {"asdg": 235, "Asg": "asg"}, "asgdagwe": "asgdouvjknmwefasdvsaivdljnm,efsdvlnk", "saegas": "asrgiufhkjnaseodigfun", "sargasdba": {"Asdg": "segaserhbaewh", "segaseh": "aehsarjt"}}
# for i in range(17):
#     recursive_obj[f"self{i}"] = copy.deepcopy(recursive_obj)
# objects.append(recursive_obj)


# objects.clear()
with open("/home/aharon/Downloads/twitter.json") as f:
    objects.append(orjson.loads(f.read()))

with open("/home/aharon/Downloads/citm_catalog.json") as f:
    objects.append(orjson.loads(f.read()))

with open("/home/aharon/Downloads/canada.json") as f:
    objects.append(orjson.loads(f.read()))



def profile(name: str, func: callable):
    tracemalloc.start()
    start = time.perf_counter()
    try:
        res = func()
        try:
            end_len = len(res)
        except TypeError:
            end_len = 0
        elapsed = (time.perf_counter() - start) * 1000
        c, peak = tracemalloc.get_traced_memory()
    except Exception as e:
        print(e)
        # raise e
        return
    finally:
        tracemalloc.stop()
    return {"name": name, "elapsed": elapsed, "end_len": end_len, "peak_memory": peak}


for obj in objects:
    print()
    serialized = pypinch.dump_bytes(obj, use_pointers=True)
    # print(serialized)
    # print(dump_bytes(obj))
    unserialized = pypinch.load_bytes(serialized, modify_input=False)
    print()
    print(str(obj)[:120])
    print(str(unserialized)[:120])
    assert (isinstance(obj, float) and math.isnan(obj) and math.isnan(unserialized)) or unserialized == obj

    # serialized = dump_bytes(obj, settings=Settings(use_pointers=True))
    # unserialized = load_bytes(serialized, settings=LoadSettings(modify_input=True))
    # assert (isinstance(obj, float) and math.isnan(obj) and math.isnan(unserialized)) or unserialized == obj
# exit()

for obj in objects:
    print(str(obj)[:120])
    results = [
        profile("pypinch (python)", lambda: pinch_py_dump(obj)),
        profile("pypinch (rust)", lambda: pinch_rust_dump(obj)),
        profile("mine w pointer & str keys (python)",
                lambda: pinch_py_dump(obj, use_pointers=True, allow_non_string_keys=False)),
        profile("pypinch w pointers (rust)", lambda: pinch_rust_dump(obj, use_pointers=True)),
        profile("json", lambda: json.dumps(obj).encode()),
        profile("orjson", lambda: orjson.dumps(obj)),
        profile("pickle", lambda: pickle.dumps(obj)),
        profile("msgpack", lambda: msgpack.dumps(obj))
    ]
    # results.append(profile("str keys", lambda: dump_bytes(obj, settings=Settings(allow_non_string_keys=False))))
    results = [x for x in results if x]

    display_benchmark(
        results,
        metrics={
            "elapsed": {
                "label": "Elapsed (ms)",
                "higher_is_better": False,
                "format": "{:.2f}",
            },
            "end_len": {
                "label": "End Len",
                "higher_is_better": False,
                "format": "{:.1f}",
            },
            "peak_memory": {
                "label": "Peak Mem",
                "higher_is_better": False,
                "format": "{:.1f}",
            },
        },
    )
    # profile("msgpack", lambda: msgpack.dumps(obj))
print("""
#################################################################################3
                                        LOADS
#################################################################################3
""")
for obj in objects:
    print(str(obj)[:120])
    mine = pinch_py_dump(obj)
    mine2 = copy.deepcopy(mine)
    mine_bytes = bytes(mine)
    mine_w_pointer = pinch_py_dump(obj, use_pointers=True)
    try:
        json_serialized = orjson.dumps(obj)
    except:
        json_serialized = None
    pickle_serialized = pickle.dumps(obj)
    try:
        msgpack_serialized = msgpack.dumps(obj)
    except:
        msgpack_serialized = None
    results = [
        profile("mine (py)", lambda: pinch_py_loads(mine, modify_input=True, use_tuples=True)),
        profile("mine (rst)", lambda: pinch_rust_loads(mine2, modify_input=True, use_tuples=True)),
        profile("mine from bytes (py)", lambda: pinch_py_loads(mine_bytes, modify_input=False, use_tuples=True)),
        profile("mine from bytes (rust)", lambda: pinch_rust_loads(mine_bytes, modify_input=False, use_tuples=True)),
        *([profile("json", lambda: json.loads(json_serialized))] if json_serialized else []),
        *([profile("orjson", lambda: orjson.loads(json_serialized))] if json_serialized else []),
        profile("pickle", lambda: pickle.loads(pickle_serialized)),
        *([profile("msgpack", lambda: msgpack.loads(msgpack_serialized))] if msgpack_serialized else [])
    ]
    results = [x for x in results if x]

    display_benchmark(
        results,
        metrics={
            "elapsed": {
                "label": "Elapsed (ms)",
                "higher_is_better": False,
                "format": "{:.2f}",
            },
            "peak_memory": {
                "label": "Peak Mem",
                "higher_is_better": False,
                "format": "{:.1f}",
            },
        },
    )
    # profile("msgpack", lambda: msgpack.dumps(obj))



import pypinch
import os


def test_backend_detection():
    force_py = os.environ.get("PYPINCH_FORCE_PYTHON") == "1"
    if force_py:
        assert pypinch._BACKEND == "python"
    else:
        assert pypinch._BACKEND == "rust"



from dataclasses import dataclass
from typing import List


@dataclass
class Settings:
    use_tuples: bool
    pointers: List[str]     # TODO: for small strings, will they be saved in her multiple times?



import copy
import datetime
import math
import sys

if sys.version_info.major == 3 and sys.version_info.minor < 9:
    from backports.zoneinfo import ZoneInfo
else:
    from zoneinfo import ZoneInfo

import pytest

import pypinch

ALL_TYPES_OF_OBJECTS = [
    (1231,),
    (1,),
    (332634,),
    (5437890568343289547384,),
    (0,),
    (-1231,),
    (-1,),
    (-332634,),
    (-5437890568343289547384,),

    (1.000000000001,),
    (100000000000000000000000000000000000000000.1,),
    (23423523.543262346234,),
    (4.4,),
    (-1.000000000001,),
    (-100000000000000000000000000000000000000000.1,),
    (-23423523.543262346234,),
    (-4.4,),

    (math.inf,),
    (-math.inf,),

    ("afsag",),
    pytest.param("092u384oiwjrklsgmfoisgjldkxfmoweij;lksgzwaoi;elgjskznwoi;jetlaksfdnv" * 1_000_000, id="long string"),
    pytest.param("".join(chr(i) for i in range(10000)), id="lots of unicode"),
    ("",),

    (b"1234",),
    (b"abcdefghijklmnopqrstuvwxyz",),
    (b"",),
    pytest.param(b"".join(bytes(i) for i in range(10000)), id="lots of bytes"),

    (None,),
    (True,),
    (False,),

    ([None] * 10,),
    ([b"1234", b"asgsa", b"sgaeg4we"],),
    ([0.1, 0.2, 0.3, 0.4],),
    ([-91, 0, 1, 2, 3, 4, 5, 6, 7, 8],),
    (list(range(50, 1000)),),
    (["aaaa", "aaaa", "aaaa"],),
    ([1, "asdg", b"234sa", 4.5, [1, 2, 3, 4, 5], False, [], None],),

    ({"a": "sdgaeiogn", "waegw": 123, "sdagweg": list(range(10)), "aegsag": {"asdg": 235, "Asg": b"asg"}},),
    ({"a": "sdgaeiogn", 1: "123"},),
    ({"": "", "\"": None},),
    ({1: "afdbda", "ar": "23wesd", False: 23453, 1234: 12324356, "": {"sgdfn32rwefsdvre": 34}},),

    ({"a": "sdgaeiogn", "content": b"1243567" * 1024 * 1024 * 50, "sdagweg": list(range(10)),
      "aegsag": {"asdg": 235, "Asg": b"asg"}},),
    pytest.param([True, False, False] * 1000, id="list of booleans"),
]


@pytest.mark.parametrize(
    ["obj"],
    ALL_TYPES_OF_OBJECTS
)
def test__serialize_deserialize__default(obj):
    # Arrange
    original_obj = copy.deepcopy(obj)

    # Act
    # print("!")
    serialized = pypinch.dump_bytes(obj)
    unserialized = pypinch.load_bytes(serialized)

    # Assert
    assert unserialized == original_obj


def test__serialize_deserialize__nan():
    # Act
    serialized = pypinch.dump_bytes(float("nan"))
    unserialized = pypinch.load_bytes(serialized)

    # Assert
    assert math.isnan(unserialized)

@pytest.mark.parametrize(
    ["input_tuple", "expected"],
    [
        (tuple(), []),
        ((1, 2, 3), [1, 2, 3]),
        ((((),),), [[[]]]),
        ((1, None, 2.3, "rtjg", b"5y4rthf", [], {}, tuple()), [1, None, 2.3, "rtjg", b"5y4rthf", [], {}, []]),
    ]
)
def test__tuples_serialize_deserialize__into_list(input_tuple, expected):
    # Act
    serialized = pypinch.dump_bytes(float("nan"))
    unserialized = pypinch.load_bytes(serialized)

    # Assert
    assert math.isnan(unserialized)

# @pytest.mark.parametrize(
#     ["obj", "expected"],
#     [
#         (datetime.datetime(2026, 10, 4, 23, 2, 9, 53, tzinfo=datetime.timezone.utc), "2026-10-04T23:02:09.000053+00:00"),
#         (
#             [
#                 datetime.datetime(2026, 10, 4, 23, 2, 9, 53, tzinfo=datetime.timezone.utc),
#                 datetime.datetime(1995, 1, 2, 6, 3, 18, tzinfo=ZoneInfo("America/Los_Angeles")),
#                 datetime.datetime(2050, 4, 1, tzinfo=ZoneInfo("Asia/Kolkata")),
#             ],
#             [
#                 "2026-10-04T23:02:09.000053+00:00",
#                 "1995-01-02T06:03:18-08:00",
#                 "2050-04-01T00:00:00+05:30",
#             ]
#         ),
#     ]
# )
# def test__serialize_unknown_types(obj, expected):
#     # Act
#     serialized = pypinch.dump_bytes(obj)
#     unserialized = pypinch.load_bytes(serialized)
#
#     # Assert
#     assert unserialized == expected


# @pytest.mark.parametrize(
#     ["obj", "expected"],
#     [
#         ([1, 2, 3], (1, 2, 3)),
#         ({"a": [], "b": [[1], ["f"]]}, {"a": tuple(), "b": ((1,), ("f",))}),
#     ]
# )
# def test__serialize_deserialize__use_tuples(obj, expected):
#     # Act
#     serialized = pypinch.dump_bytes(obj)
#     unserialized = pypinch.load_bytes(serialized, use_tuples=True, modify_input=True)
#
#     # Assert
#     assert unserialized == expected



from pypinch.consts import NUMBER_BASE, ENDING_FLAG


# TODO optimize how negative ints are stored (only the first digit needs to be in a different base)
def encode_number(buffer: bytearray, num: int, base: int = NUMBER_BASE) -> None:
    if num < base:
        buffer.append(num)
    else:
        buffer.append(ENDING_FLAG)
        num -= base
        while num:
            num, remainder = divmod(num, base)
            buffer.append(remainder)
        buffer.append(ENDING_FLAG)



import os

FORCE_PYTHON = os.environ.get("PYPINCH_FORCE_PYTHON")

_pypinch = None
_BACKEND = None
if not FORCE_PYTHON:
    try:
        from ._pypinch import *
    except ImportError:
        _pypinch = None

if _pypinch is not None:
    _BACKEND = "rust"
    dump_bytes = _pypinch.dump_bytes
    load_bytes = _pypinch.load_bytes
else:
    _BACKEND = "python"
    from .serialize.serialize import dump_bytes
    from .deserialize.deserialize import load_bytes

pinch = dump_bytes
unpinch = load_bytes



import pytest

from pypinch.exceptions import DeserializationError, SerializationError
import pypinch


SERIALIZED = pypinch.dump_bytes(
    [{}, {"a": "1", "b": 2}, {1: "a"}, [1, 2, 3, 4], [True, False] * 10, [1, True, b"Aag" * 100], 1234567865432135456857435241356878576435241321435465879586435241312345678654321354568574352413568785764352413214354658795864352413]
)


@pytest.mark.parametrize(
    ["data"],
    [
        (SERIALIZED[:i],)
        for i in range(len(SERIALIZED))
    ]
)
def test__load_invalid_data__dies_gracefully(data: bytes):
    try:
        pypinch.load_bytes(data)
    except DeserializationError:
        assert True
    except:
        assert False


def test__dump_dict_w_tuple_key__serialization_error():
    try:
        pypinch.dump_bytes({(1, 2): 4})
    except SerializationError:
        assert True
    except:
        assert False
    else:
        assert False



from typing import Tuple

from pypinch.consts import ByteLike, NUMBER_BASE, ENDING_FLAG


def decode_number(buffer: ByteLike, pointer: int, base: int = NUMBER_BASE) -> Tuple[int, int]:
    """ numbers are stored in this format:
    if the number is smaller than ENDING_FLAG:
        just store the number (takes 1 byte), this is most cases so this is the most optimized
    if the number is bigger than ENDING_FLAG:
        first store an ENDING_FLAG (to differentiate from the numbers which are smaller than ENDING_FLAG)
        now, we know the number is at least `base` (so add that to the result)
        now store the number in base `base` up until we reach an ENDING_FLAG which signals the number is over
    """
    if buffer[pointer] != ENDING_FLAG:
        return buffer[pointer], pointer + 1
    power = 1
    res = base
    pointer += 1
    while buffer[pointer] != ENDING_FLAG:
        res += buffer[pointer] * power
        power *= base
        pointer += 1
    return res, pointer + 1



from rich.console import Console
from rich.table import Table
from rich.text import Text


def display_benchmark(
    results,
    *,
    name_key="name",
    metrics=None,
    sort_by=None,
):
    """
    results: list[dict]
        Example:
        [
            {"name": "func_a", "time_ms": 12.3, "mem_mb": 45},
            ...
        ]

    metrics: dict[str, dict]
        Metric configuration:
        {
            "time_ms": {
                "label": "Time (ms)",
                "higher_is_better": False,
                "format": "{:.2f}",
            },
            "mem_mb": {
                "label": "Memory (MB)",
                "higher_is_better": False,
                "format": "{:.1f}",
            },
        }

    sort_by: str | None
        Which metric to sort by
    """

    if not results:
        return

    # console = Console()
    console = Console(force_terminal=True, color_system="truecolor")

    # Auto infer metrics if not supplied
    if metrics is None:
        sample = results[0]
        metrics = {
            k: {
                "label": k,
                "higher_is_better": False,
                "format": "{}",
            }
            for k in sample
            if k != name_key
        }

    # Gather per-metric ranges
    ranges = {}
    for metric in metrics:
        values = [r[metric] for r in results if metric in r]
        ranges[metric] = (min(values), max(values))

    def color_scale(value, vmin, vmax, higher_is_better):
        if vmax == vmin:
            return "white"

        ratio = (value - vmin) / (vmax - vmin)

        # If lower is better → invert scale
        if higher_is_better:
            ratio = 1 - ratio

        if ratio < 0.33:
            return "green"
        elif ratio < 0.66:
            return "yellow"
        return "red"

    # Sorting
    if sort_by:
        higher_is_better = metrics[sort_by].get("higher_is_better", False)
        results = sorted(
            results,
            key=lambda r: r[sort_by],
            reverse=higher_is_better,
        )

    # Build table
    table = Table(header_style="bold")
    table.add_column(name_key.capitalize(), style="bold")

    for metric, cfg in metrics.items():
        table.add_column(cfg.get("label", metric), justify="right")

    # Populate rows
    for r in results:
        row = [str(r[name_key])]

        for metric, cfg in metrics.items():
            value = r[metric]
            vmin, vmax = ranges[metric]

            color = color_scale(
                value,
                vmin,
                vmax,
                cfg.get("higher_is_better", False),
            )

            fmt = cfg.get("format", "{}")
            row.append(Text(fmt.format(value), style=color))

        table.add_row(*row)

    console.print(table)



import gc
import struct
import typing
from typing import Tuple, List

from pypinch.consts import NUMBER_BASE, ObjType, POSITIVE_INT_FLAG, NULL_FLAG, BYTES_FLAG, \
    LIST_FLAG, \
    DICT_FLAG, STR_KEY_DICT_FLAG, FLOAT_FLAG, STR_FLAG, NEGATIVE_INT_FLAG, \
    EMPTY_LIST_FLAG, EMPTY_DICT_FLAG, CONSISTENT_TYPE_LIST_FLAG, INT_FLAG, BOOL_FLAG, POINTER_FLAG, \
    ByteLike, HEADER, BIG_ENDIAN_DOUBLE_FORMAT, NUMBER_OF_BITS_IN_BYTE, \
    LEFTMOST_BIT_MASK, BYTES_IN_DOUBLE, FIRST_FLAGS_LIST, AMOUNT_OF_USED_FLAGS, \
    INVALID_UTF_8_START_BYTE_COMPACT_ASCII, ASCII_STR_FLAG, LIST_OF_STRUCTURED_DICTS_FLAG

from pypinch.exceptions import DeserializationError
from pypinch.deserialize.settings import Settings
from pypinch.deserialize.utils import decode_number


def load_bytes(
    buffer: ByteLike,
    *,
    use_tuples: bool = False,
    stop_gc: bool = False,
    ignore_extra_data: bool = False
) -> ObjType:

    try:
        if stop_gc:
            gc.freeze()

        settings = Settings(
            use_tuples=use_tuples,  # TODO
            pointers=[],
        )
        res, pointer = deserialize_object(buffer, len(HEADER), settings)
        if not ignore_extra_data and pointer != len(buffer):
            raise DeserializationError("Extra data")
        return res
    except DeserializationError:
        raise
    except MemoryError:
        raise
    except Exception as e:
        raise DeserializationError() from e
    finally:
        if stop_gc:
            gc.unfreeze()


def deserialize_object(buffer: bytes, pointer: int, settings: Settings) -> (ObjType, int):
    flag = buffer[pointer]
    pointer += 1
    if flag < len(FIRST_FLAGS_LIST):
        return FIRST_FLAGS_LIST[flag], pointer
    elif flag == POSITIVE_INT_FLAG:
        return decode_number(buffer, pointer)
    elif flag == STR_KEY_DICT_FLAG:
        length, pointer = decode_number(buffer, pointer)
        res_dict = {}
        for i in range(length):
            if buffer[pointer] == NUMBER_BASE - 1:
                position, pointer = decode_number(buffer, pointer + 1)
                k = settings.pointers[position]
            else:
                k, pointer = deserialize_str(buffer, pointer, settings, base=NUMBER_BASE - 1)
            v, pointer = deserialize_object(buffer, pointer, settings)
            res_dict[k] = v
        return res_dict, pointer
    elif flag == ASCII_STR_FLAG:
        return deserialize_str(buffer, pointer, settings)
    elif flag == STR_FLAG:
        return deserialize_str(buffer, pointer, settings)
    elif flag == DICT_FLAG:
        length, pointer = decode_number(buffer, pointer)
        res_dict = {}
        for _ in range(length):
            if buffer[pointer] == STR_FLAG:
                # fast path
                k, pointer = deserialize_str(buffer, pointer + 1, settings)
            else:
                k, pointer = deserialize_object(buffer, pointer, settings)
            v, pointer = deserialize_object(buffer, pointer, settings)
            res_dict[k] = v
        return res_dict, pointer
    elif flag == EMPTY_DICT_FLAG:
        return {}, pointer
    elif flag == LIST_FLAG:
        length, pointer = decode_number(buffer, pointer)
        res_list = [None] * length
        for i in range(length):
            res_list[i], pointer = deserialize_object(buffer, pointer, settings)
        return res_list, pointer
    elif flag == EMPTY_LIST_FLAG:
        return (tuple() if settings.use_tuples else []), pointer
    elif flag == CONSISTENT_TYPE_LIST_FLAG:
        typ_flag = buffer[pointer]
        length, pointer = decode_number(buffer, pointer + 1)
        if typ_flag == NULL_FLAG:
            return ((None,) if settings.use_tuples else [None]) * length, pointer
        elif typ_flag == INT_FLAG:
            res_list = typing.cast(List[int], [None] * length)
            for i in range(length):
                if buffer[pointer] == NUMBER_BASE - 1:
                    num, pointer = decode_number(buffer, pointer + 1, base=NUMBER_BASE - 1)
                    res_list[i] = -num
                else:
                    num, pointer = decode_number(buffer, pointer, base=NUMBER_BASE - 1)
                    res_list[i] = num
            return res_list, pointer
        elif typ_flag == BOOL_FLAG:
            res_list = typing.cast(List[bool], [None] * length)
            # same as: math.ceil(length / NUMBER_OF_BITS_IN_BYTE)
            # the `>> 3` is like dividing by 8 (8 is `1000` in binary)
            # the + 7 is like rounding up
            length_in_bytes = (length + 7) >> 3
            try:
                for i in range(length_in_bytes):
                    byte = buffer[pointer + i]
                    for j in range(NUMBER_OF_BITS_IN_BYTE):
                        res_list[i * NUMBER_OF_BITS_IN_BYTE + j] = (byte & LEFTMOST_BIT_MASK) == LEFTMOST_BIT_MASK
                        byte <<= 1
            except IndexError:
                pass
            return res_list, pointer + length_in_bytes
        elif typ_flag == BYTES_FLAG:
            res_list = typing.cast(List[bytes], [None] * length)
            for i in range(length):
                bytes_length, pointer = decode_number(buffer, pointer)
                res_list[i] = bytes(buffer[pointer:pointer + bytes_length])
                pointer += bytes_length
            return res_list, pointer
        elif typ_flag == STR_FLAG:
            res_list = typing.cast(List[str], [None] * length)
            for i in range(length):
                res_list[i], pointer = deserialize_str(buffer, pointer, settings)
            return res_list, pointer
        elif typ_flag == FLOAT_FLAG:
            res_list = [
                struct.unpack(BIG_ENDIAN_DOUBLE_FORMAT, buffer[pos:pos + BYTES_IN_DOUBLE])[0]
                for pos in range(pointer, pointer + BYTES_IN_DOUBLE * length, BYTES_IN_DOUBLE)
            ]
            pointer += BYTES_IN_DOUBLE * length
            return res_list, pointer
        else:
            raise DeserializationError(f"Unexpected type flag: {typ_flag}")
    elif flag == NEGATIVE_INT_FLAG:
        num, pointer = decode_number(buffer, pointer)
        return -num, pointer
    elif flag == FLOAT_FLAG:
        num = struct.unpack(BIG_ENDIAN_DOUBLE_FORMAT, buffer[pointer:pointer + BYTES_IN_DOUBLE])[0]
        return num, pointer + BYTES_IN_DOUBLE
    elif flag == BYTES_FLAG:
        length, pointer = decode_number(buffer, pointer)
        return bytes(buffer[pointer:pointer + length]), pointer + length
    elif flag == POINTER_FLAG:
        position, pointer = decode_number(buffer, pointer)
        return settings.pointers[position], pointer
    elif flag == LIST_OF_STRUCTURED_DICTS_FLAG:
        list_length, pointer = decode_number(buffer, pointer)
        dict_length, pointer = decode_number(buffer, pointer)
        res_list = typing.cast(List[typing.Dict], [None] * list_length)
        # first dict:
        first_dict = {}
        keys = typing.cast(List[str], [None] * dict_length)
        for i in range(dict_length):
            k, pointer = deserialize_object(buffer, pointer, settings)
            v, pointer = deserialize_object(buffer, pointer, settings)
            first_dict[k] = v
            keys[i] = k
        res_list[0] = first_dict
        # rest of the dicts:
        for list_idx in range(1, list_length):
            dct = {}
            for i, key in enumerate(keys):
                dct[key], pointer = deserialize_object(buffer, pointer, settings)
            res_list[list_idx] = dct
        return res_list, pointer
    elif flag < AMOUNT_OF_USED_FLAGS:
        raise DeserializationError("unexpected flag")
    else:
        return flag - AMOUNT_OF_USED_FLAGS, pointer


def deserialize_str(buffer: bytes, pointer: int, settings: Settings, base: int = NUMBER_BASE) -> Tuple[str, int]:
    length, pointer = decode_number(buffer, pointer, base=base)
    string = buffer[
        #          Skip 1 char if buffer starts with INVALID_UTF_8_START_BYTE_COMPACT_ASCII
        pointer + (buffer[pointer] == INVALID_UTF_8_START_BYTE_COMPACT_ASCII)
        :pointer + length
    ].decode()
    settings.pointers.append(string)
    return string, pointer + length



import bisect
import struct
from datetime import datetime
from typing import Union, List, Tuple

from pypinch.consts import NUMBER_BASE, ObjType, POSITIVE_INT_FLAG, FALSE_FLAG, TRUE_FLAG, NULL_FLAG, BYTES_FLAG, \
    LIST_FLAG, \
    DICT_FLAG, STR_KEY_DICT_FLAG, FLOAT_FLAG, STR_FLAG, NEGATIVE_INT_FLAG, EMPTY_STR_FLAG, EMPTY_BYTES_FLAG, \
    EMPTY_LIST_FLAG, EMPTY_DICT_FLAG, AMOUNT_OF_USED_FLAGS, CONSISTENT_TYPE_LIST_FLAG, INT_FLAG, BOOL_FLAG, \
    POINTER_FLAG, HEADER, \
    BIG_ENDIAN_DOUBLE_FORMAT, NUMBER_OF_BITS_IN_BYTE, ENCODED_NUMBER_LIMITS, \
    ASCII_STR_FLAG, INVALID_UTF_8_START_BYTE_COMPACT_ASCII, LIST_OF_STRUCTURED_DICTS_FLAG
from pypinch.exceptions import SerializationError
from pypinch.serialize.settings import Settings
from pypinch.serialize.utils import encode_number

_pack_double = struct.Struct(BIG_ENDIAN_DOUBLE_FORMAT).pack


def dump_bytes(obj: ObjType, *, allow_non_string_keys: bool = True, serialize_dates: bool = False) -> bytes:
    try:
        settings = Settings(
            allow_non_string_keys=allow_non_string_keys,
            pointers={},
            serialize_dates=serialize_dates,
            str_count=0
        )
        buffer = bytearray(HEADER)
        serialize_object_with_type(buffer, obj, settings)
        return bytes(buffer)
    except SerializationError:
        raise
    except MemoryError:
        raise
    except Exception as e:
        raise SerializationError() from e


def serialize_object_with_type(buffer: bytearray, obj: ObjType, settings: Settings) -> None:
    typ = type(obj)
    if typ is str:
        if len(obj) == 0:
            buffer.append(EMPTY_STR_FLAG)
            return
        if (prev_pos := settings.pointers.get(obj)) is not None:
            predicted_pointer_length = bisect.bisect_left(ENCODED_NUMBER_LIMITS, prev_pos)
            predicted_str_length = len(obj) + bisect.bisect_left(ENCODED_NUMBER_LIMITS, len(obj))
            if predicted_pointer_length <= predicted_str_length:
                buffer.append(POINTER_FLAG)
                encode_number(buffer, prev_pos)
                return
        else:
            settings.pointers[obj] = settings.str_count
        settings.str_count += 1
        try:
            encoded_str = obj.encode(encoding="ascii")
            buffer.append(ASCII_STR_FLAG)
        except UnicodeEncodeError:
            buffer.append(STR_FLAG)
            encoded_str = obj.encode()
        encode_number(buffer, len(encoded_str))
        buffer.extend(encoded_str)
    elif typ is int:
        if obj >= 0:
            if obj < NUMBER_BASE - AMOUNT_OF_USED_FLAGS:
                buffer.append(AMOUNT_OF_USED_FLAGS + obj)
            else:
                buffer.append(POSITIVE_INT_FLAG)
                encode_number(buffer, obj)
        else:
            buffer.append(NEGATIVE_INT_FLAG)
            encode_number(buffer, -obj)
    elif typ is bool:
        buffer.append(TRUE_FLAG if obj else FALSE_FLAG)
    elif obj is None:
        buffer.append(NULL_FLAG)
    elif typ is list or typ is tuple:
        if len(obj) == 0:
            buffer.append(EMPTY_LIST_FLAG)
        elif is_consistent_type_list(obj):
            first_type = type(obj[0])
            if obj[0] is None:
                buffer.append(CONSISTENT_TYPE_LIST_FLAG)
                buffer.append(NULL_FLAG)
                encode_number(buffer, len(obj))
            elif first_type is int:
                # no longer have the flag to distinguish between positive and negative numbers so do this instead
                buffer.append(CONSISTENT_TYPE_LIST_FLAG)
                buffer.append(INT_FLAG)
                encode_number(buffer, len(obj))
                for item in obj:
                    if item < 0:
                        buffer.append(NUMBER_BASE - 1)
                        encode_number(buffer, -item, base=NUMBER_BASE - 1)
                    else:
                        encode_number(buffer, item, base=NUMBER_BASE - 1)
            elif first_type is bool:
                buffer.append(CONSISTENT_TYPE_LIST_FLAG)
                buffer.append(BOOL_FLAG)
                encode_number(buffer, len(obj))
                byte = number_of_bits = 0
                for item in obj:
                    byte = (byte << 1) | item
                    number_of_bits += 1
                    if number_of_bits == NUMBER_OF_BITS_IN_BYTE:
                        buffer.append(byte)
                        byte = number_of_bits = 0
                if number_of_bits:
                    buffer.append(byte << (NUMBER_OF_BITS_IN_BYTE - number_of_bits))
            elif first_type is dict:
                first_keys = obj[0].keys()
                if (
                    all(type(k) is str for k in first_keys)
                    and all(len(x) == len(first_keys) for x in obj[1:])
                    and all(x.keys() == first_keys for x in obj[1:])
                ):
                    buffer.append(LIST_OF_STRUCTURED_DICTS_FLAG)
                    encode_number(buffer, len(obj))
                    encode_number(buffer, len(first_keys))

                    # first dict:
                    for k, v in obj[0].items():
                        serialize_object_with_type(buffer, k, settings)
                        serialize_object_with_type(buffer, v, settings)

                    # the rest:
                    for item in obj[1:]:
                        for key in first_keys:
                            serialize_object_with_type(buffer, item[key], settings)
                else:
                    serialize_normal_list(buffer, obj, settings)
            else:
                buffer.append(CONSISTENT_TYPE_LIST_FLAG)
                try:
                    # todo: support str (with utf trick)?
                    buffer.append({bytes: BYTES_FLAG, float: FLOAT_FLAG, datetime: STR_FLAG}[first_type])
                except KeyError:
                    raise SerializationError(f"Unexpected type: {first_type}")

                encode_number(buffer, len(obj))
                for item in obj:
                    serialize_object_without_type(buffer, item, settings)
        else:
            serialize_normal_list(buffer, obj, settings)
    elif typ is dict:
        if len(obj) == 0:
            buffer.append(EMPTY_DICT_FLAG)
        # TODO: on lists as well and in serialize_without_type
        elif all(type(x) is str for x in obj.keys()):
            buffer.append(STR_KEY_DICT_FLAG)
            encode_number(buffer, len(obj))
            for k, v in obj.items():
                if (prev_pos := settings.pointers.get(k)) is not None:
                    predicted_digits = 1 + bisect.bisect_left(ENCODED_NUMBER_LIMITS, prev_pos)
                    if predicted_digits <= len(obj):
                        buffer.append(NUMBER_BASE - 1)
                        encode_number(buffer, prev_pos)
                    else:
                        serialize_str_without_type(buffer, k, settings, base=NUMBER_BASE - 1)
                else:
                    serialize_str_without_type(buffer, k, settings, base=NUMBER_BASE - 1)
                serialize_object_with_type(buffer, v, settings)
        else:
            buffer.append(DICT_FLAG)
            encode_number(buffer, len(obj))
            for k, v in obj.items():
                if type(k) is tuple:
                    raise SerializationError("Invalid type for dict key: tuple")
                serialize_object_with_type(buffer, k, settings)
                serialize_object_with_type(buffer, v, settings)
    elif typ is float:
        buffer.append(FLOAT_FLAG)
        buffer.extend(_pack_double(obj))
    elif typ is bytes:
        if len(obj) == 0:
            buffer.append(EMPTY_BYTES_FLAG)
        else:
            buffer.append(BYTES_FLAG)
            encode_number(buffer, len(obj))
            buffer.extend(obj)
    elif typ is datetime and settings.serialize_dates:
        return serialize_object_with_type(buffer, obj.isoformat(), settings)
    else:
        if typ is datetime and not settings.serialize_dates:
            raise SerializationError(f"Unexpected type: datetime, with flag serialize_dates disabled")
        raise SerializationError(f"Unexpected type: {typ}")


def serialize_normal_list(buffer: bytearray, obj: Union[List, Tuple], settings: Settings) -> None:
    buffer.append(LIST_FLAG)
    encode_number(buffer, len(obj))
    for item in obj:
        serialize_object_with_type(buffer, item, settings)


def is_consistent_type_list(obj: Union[List, Tuple]) -> bool:
    if len(obj) <= 1:
        return False
    first_type = type(obj[0])
    if first_type in [list, tuple, str]:
        return False
    return all(type(x) is first_type for x in obj)


def serialize_object_without_type(buffer: bytearray, obj: ObjType, settings: Settings) -> None:
    typ = type(obj)
    if typ is int:
        encode_number(buffer, obj if obj > 0 else -obj)
    elif typ is bool:
        buffer.append(TRUE_FLAG if obj else FALSE_FLAG)
    elif obj is None:
        buffer.append(NULL_FLAG)
    elif typ is bytes:
        encode_number(buffer, len(obj))
        buffer.extend(obj)
    elif typ is list or typ is tuple:
        encode_number(buffer, len(obj))
        for item in obj:
            serialize_object_with_type(buffer, item, settings)
    elif typ is dict:
        if len(obj) == 0:
            buffer.append(EMPTY_DICT_FLAG)
        # elif not settings.use_pointers and not settings.allow_non_string_keys:
        #     buffer.append(STR_KEY_DICT_FLAG)
        #     encode_number(buffer, len(obj))
        #     for k, v in obj.items():
        #         if type(k) is not str:
        #             raise EncodingError("Encountered a non string key while allow_non_string_keys is False")
        #         serialize_object_without_type(buffer, k, settings)
        #         serialize_object_with_type(buffer, v, settings)
        # elif not settings.use_pointers and all(type(x) is str for x in obj.keys()):
        # TODO:    buffer.append(STR_KEY_DICT_FLAG)
        #     encode_number(buffer, len(obj))
        #     for k, v in obj.items():
        #         serialize_object_without_type(buffer, k, settings)
        #         serialize_object_with_type(buffer, v, settings)
        else:
            buffer.append(DICT_FLAG)
            encode_number(buffer, len(obj))
            for k, v in obj.items():
                if type(k) is tuple:
                    raise SerializationError("Invalid type for dict key: tuple")
                serialize_object_with_type(buffer, k, settings)
                serialize_object_with_type(buffer, v, settings)
    elif typ is float:
        buffer.extend(_pack_double(obj))
    elif typ is str:
        serialize_str_without_type(buffer, obj, settings)
    elif typ is datetime and settings.serialize_dates:
        return serialize_object_without_type(buffer, obj.isoformat(), settings)
    else:
        raise SerializationError(f"Unexpected type: {typ}")


def serialize_str_without_type(buffer: bytearray, obj: ObjType, settings: Settings, base: int = NUMBER_BASE) -> None:
    try:
        encoded_str = obj.encode(encoding="ascii")
        encode_number(buffer, 1 + len(encoded_str), base=base)
        buffer.append(INVALID_UTF_8_START_BYTE_COMPACT_ASCII)
    except UnicodeEncodeError:
        encoded_str = obj.encode()
        encode_number(buffer, len(encoded_str), base=base)
    settings.pointers[obj] = settings.str_count
    settings.str_count += 1
    buffer.extend(encoded_str)
